﻿using Phidget22;
using Phidget22.Events;
using Phidget22.ExampleUtils;
using System;
using System.Drawing;
//using System.Threading.Tasks;
using System.Windows.Forms;

namespace Encoder_Example {
	public partial class Form1 : Form {
		CommandLineOpen openArgs;
		Phidget22.Encoder openCh = null;
		Phidget22.Encoder encoder = null;
		formCleaner cleaner;
		private ErrorEventBox errorBox;
		int CPR = -1;
		double prevVelocity = 0;
		double prevTime = 0;

		private DataGraphBox positionGraphBox, velocityGraphBox, calcPositionGraphBox;

		FastEventHandler<EncoderPositionChangeEventArgs> encoderPositionChangeEventHandler;

		public Form1() {
			InitializeComponent();
			openArgs = new CommandLineOpen(this);
			cleaner = new formCleaner(this);
		}

		public Form1(Phidget22.Encoder ch) {
			InitializeComponent();
			openCh = ch;
			cleaner = new formCleaner(this);
		}

		private void Form1_Load(object sender, EventArgs e) {
			phidgetInfoBox1.Clear();

			errorBox = new ErrorEventBox(this);
			encoder = openCh ?? openArgs.makeChannel<Phidget22.Encoder>();

			positionGraphBox = new DataGraphBox(this, "Position");
			velocityGraphBox = new DataGraphBox(this, "Velocity");
			calcPositionGraphBox = new DataGraphBox(this, "Angle");

			encoder.Attach += enc_attach;
			encoder.Detach += enc_detach;
			encoder.Error += enc_error;
			encoder.PropertyChange += enc_propertychange;

			// This causes the events to get handled without invoke
			encoderPositionChangeEventHandler = new FastEventHandler<EncoderPositionChangeEventArgs>();
			encoder.PositionChange += encoderPositionChangeEventHandler.Handler;
			encoderPositionChangeEventHandler.Event += enc_change;

			try {
				encoder.Open();
			} catch (PhidgetException ex) { errorBox.addMessage("Error opening device: " + ex.Message); }
		}

		private void Form1_FormClosing(object sender, FormClosingEventArgs e) {

			encoder.Attach -= enc_attach;
			encoder.Detach -= enc_detach;
			encoder.Error -= enc_error;
			encoder.PositionChange -= encoderPositionChangeEventHandler.Handler;
			encoder.PropertyChange -= enc_propertychange;
			encoder.Close();
		}
		void enc_attach(object sender, Phidget22.Events.AttachEventArgs e) {
			try {
				phidgetInfoBox1.FillPhidgetInfo((Phidget)sender);

				Phidget22.Encoder attachedDevice = (Phidget22.Encoder)sender;

				positionGraphBox.SetTitle(DataGraphBox.GenerateTitle("Position", (Phidget)encoder));
				velocityGraphBox.SetTitle(DataGraphBox.GenerateTitle("Velocity", (Phidget)encoder));
				calcPositionGraphBox.SetTitle(DataGraphBox.GenerateTitle("Angle", (Phidget)encoder));

				try {
					dataIntervalTrk.Unit = "ms";
					dataIntervalTrk.initializeRange(attachedDevice.MinDataInterval, attachedDevice.MaxDataInterval);
					dataIntervalTrk.Value = attachedDevice.DataInterval;

					//differing max/min means the interval is a settable configuration
					if (attachedDevice.MaxDataInterval == attachedDevice.MinDataInterval)
						dataIntervalTrk.Enabled = false;

					changeTriggerTrk.Unit = " pulses";
					changeTriggerTrk.trkText = changeTriggerTrk.minTxt = attachedDevice.MinPositionChangeTrigger.ToString();
					changeTriggerTrk.Minimum = attachedDevice.MinPositionChangeTrigger;
					changeTriggerTrk.Maximum = attachedDevice.MaxPositionChangeTrigger;
					changeTriggerTrk.maxTxt = attachedDevice.MaxPositionChangeTrigger.ToString();
					changeTriggerTrk.Value = attachedDevice.PositionChangeTrigger;
					changeTriggerTrk.trkText = attachedDevice.PositionChangeTrigger.ToString("F0") + changeTriggerTrk.Unit;

				} catch (PhidgetException ex) {
					errorBox.addMessage("Error setting data interval or change trigger: " + ex.Message);
				}

				try // Populate ioMode control and hide if ioMode is unsupported
				{
					EncoderIOMode io = encoder.IOMode;

					foreach (EncoderIOMode eio in Enum.GetValues(typeof(EncoderIOMode)))
						iomodeCombo.Items.Add(eio);
					iomodeCombo.SelectedIndex = 0;
					iomodePanel.Show();

				} catch (PhidgetException ex) {
					if (ex.ErrorCode == ErrorCode.Unsupported)
						iomodePanel.Hide();
				}

				_positionTxt = "N/A";
				_timeTxt = "N/A";
				_indexTxt = "N/A";
				_cumulativePosition = "N/A";
				_rpmTxt = "";
				_calcPositionTxt = "N/A";

				try {
					attachedDevice.Enabled = attachedDevice.Enabled;

					enabledChk.Visible = true;
					enabledChk.Checked = true;
				} catch (PhidgetException ex) {
					if (ex.ErrorCode == ErrorCode.Unsupported) {
						enabledChk.Visible = false;
					} else
						errorBox.addMessage("Error initializing: " + ex.Message);
				}

				settingsBox.Visible = true;
				outputBox.Visible = true;

				guiUpdateTimer.Start();
			} catch (PhidgetException ex) {
				errorBox.addMessage("Error during attach event: " + ex.Message);
			}
		}
		void enc_detach(object sender, Phidget22.Events.DetachEventArgs e) {
			guiUpdateTimer.Stop();
			_positionTxt = "";
			_timeTxt = "";
			_indexTxt = "";
			_cumulativePosition = "";
			_rpmTxt = "";
			_calcPositionTxt = "";

			phidgetInfoBox1.Clear();
			cleaner.clean();

			settingsBox.Visible = false;
			outputBox.Visible = false;

			iomodeCombo.Items.Clear();

			enabledChk.Visible = false;
			cprTxt.Text = "";
			CPR = -1;
			dataIntervalTrk.TrueValue = -1;
			rpmTxt.Text = "N/A";
			calcPositionTxt.Text = "N/A";
		}
		void enc_error(object sender, Phidget22.Events.ErrorEventArgs e) {
			errorBox.addMessage(e.Description);
		}

		private void enc_propertychange(object sender, PropertyChangeEventArgs e) {

			switch (e.PropertyName) {
			case "DataInterval":
				if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.INTERVAL) {
					dataIntervalTrk.Value = encoder.DataInterval;
					updateGraphDataInterval();
				}
				break;
			case "PositionChangeTrigger":
				changeTriggerTrk.Value = encoder.PositionChangeTrigger;
				break;
			case "DataRate":
				if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.RATE) {
					dataIntervalTrk.Value = encoder.DataRate;
					updateGraphDataInterval();
				}
				break;
			}
		}

		private string _positionTxt = "";
		private string _timeTxt = "";
		private string _indexTxt = "";
		private string _cumulativePosition = "";
		private string _rpmTxt = "";
		private string _calcPositionTxt = "";
		private void guiUpdateTimer_Tick(object sender, EventArgs e) {
			positionTxt.Text = _positionTxt;
			timeTxt.Text = _timeTxt;
			indexTxt.Text = _indexTxt;
			cumulativePosition.Text = _cumulativePosition;
			rpmTxt.Text = _rpmTxt;
			calcPositionTxt.Text = _calcPositionTxt;
		}
		void enc_change(object sender, Phidget22.Events.EncoderPositionChangeEventArgs e) {
			try {
				_positionTxt = e.PositionChange.ToString() + " pulses";
				_timeTxt = (e.TimeChange).ToString("F3") + " ms";

				if (e.IndexTriggered) {
					_indexTxt = encoder.IndexPosition.ToString() + " pulses";
				}

				_cumulativePosition = encoder.Position.ToString() + " pulses";

				double multiplier = 4;
				if (encoder.DeviceID == DeviceID.PN_1065)
					multiplier = 1;

				if (CPR != -1) {
					if (rotarySelect.Checked) {
						double timeChangeMinutes = e.TimeChange / 60000.0; // ms -> minutes
						if (prevVelocity == 0)
							timeChangeMinutes += prevTime;
						prevTime = timeChangeMinutes;
						double velocity = (((double)e.PositionChange / CPR) / timeChangeMinutes);
						_rpmTxt = (velocity / multiplier).ToString("F0") + " RPM"; // /4 to account for cpr vs ppr)

						double angle = (((double)(encoder.Position) / (multiplier * CPR)) * 360);
						_calcPositionTxt = angle.ToString() + "°";
						calcPositionGraphBox.addData(angle, encoder.DataInterval);
					} else if (linearSelect.Checked) {
						double altCPR = CPR;
						if (encoder.DeviceID == DeviceID.PN_1065)
							altCPR *= 4;
						double timeChangeSeconds = e.TimeChange / 1000.0; // ms -> minutes
						if (prevVelocity == 0)
							timeChangeSeconds += prevTime;
						prevTime = timeChangeSeconds;
						double velocity = (((double)e.PositionChange * altCPR) / (timeChangeSeconds * 1000)); // divide by 1000 to convert micron to mm
						_rpmTxt = velocity.ToString("F1") + " mm/s";


						double mm = (((double)(encoder.Position)) * ((altCPR) / 1000));
						_calcPositionTxt = mm.ToString() + " mm";
						calcPositionGraphBox.addData(mm, encoder.DataInterval);
					}


					{
						double timeChangeMinutes = e.TimeChange / 60000.0; // ms -> minutes
						if (prevVelocity == 0)
							timeChangeMinutes += prevTime;
						prevTime = timeChangeMinutes;
						double velocity = (((double)e.PositionChange / CPR) / timeChangeMinutes);

						if (e.PositionChange != 0 || timeChangeMinutes > (1 / 60.0))
							velocityGraphBox.addData(velocity, encoder.DataInterval);
						prevVelocity = velocity;
					}
				}

				positionGraphBox.addData(encoder.Position, encoder.DataInterval);
			} catch (PhidgetException ex) {
				errorBox.addMessage("Error reading position: " + ex.Message);
			}
		}

		private void enabledChk_CheckedChanged(object sender, EventArgs e) {
			try {
				encoder.Enabled = enabledChk.Checked;
			} catch (PhidgetException ex) {
				errorBox.addMessage("Error setting enabled: " + ex.Message);
			}
		}

		private void changeTriggertrk_Scroll(object sender, EventArgs e) {
			if (MouseButtons != MouseButtons.Left) {
				try {
					encoder.PositionChangeTrigger = (int)changeTriggerTrk.Value;
				} catch (Exception ex) {
					errorBox.addMessage("Error setting change trigger: " + ex.Message);
				}
			}
		}

		private void changeTriggertrk_MouseUp(object sender, MouseEventArgs e) {
			try {
				encoder.PositionChangeTrigger = (int)changeTriggerTrk.Value;
			} catch (Exception ex) {
				errorBox.addMessage("Error setting change trigger: " + ex.Message);
			}
		}

		private void cprTxt_KeyDown(object sender, KeyEventArgs e) {
			if (e.KeyCode == Keys.Enter) {
				try {
					if (cprTxt.Text != "" && cprTxt.Text != "0")
						CPR = Int32.Parse(cprTxt.Text);
					else {
						CPR = -1;
						rpmTxt.Text = "N/A";
						cprTxt.Text = "";
					}
				} catch (Exception) {
					errorBox.addMessage("Invalid CPR value.");
					cprTxt.Text = "";
				}
				e.SuppressKeyPress = true;
			}
		}

		private void CPRSetBtn_Click(object sender, EventArgs e) {
			try {
				CPR = Int32.Parse(cprTxt.Text);
			} catch (Exception) {
				errorBox.addMessage("Invalid CPR value.");
				cprTxt.Text = "";
			}
		}

		private void rstBtn_Click(object sender, EventArgs e) {
			try {
				encoder.Position = 0;
				positionGraphBox.addData(0, encoder.DataInterval);
			} catch (Exception ex) {
				errorBox.addMessage("Error setting position: " + ex.Message);
			}
		}
		private void updateGraphDataInterval() {
			try {
				if (encoder.PositionChangeTrigger == 0)
					positionGraphBox.DataInterval = (1000.0 / encoder.DataRate);
				else
					positionGraphBox.DataInterval = null;
			} catch (PhidgetException ex) { errorBox.addMessage("Error updating data interval: " + ex.Message); }
		}
		private void iomodeCombo_SelectedIndexChanged(object sender, EventArgs e) {
			try {
				encoder.IOMode = (EncoderIOMode)iomodeCombo.Items[iomodeCombo.SelectedIndex];
			} catch (Exception ex) {
				errorBox.addMessage("Error setting IO Mode: " + ex.Message);
			}
		}

		private void positionGraphButton_Click(object sender, EventArgs e) {
			if (!positionGraphBox.Visible)
				positionGraphBox.Show();
		}

		private void dataIntervalTrk_KeyOverride(object sender, KeyEventArgs e) {
			e.SuppressKeyPress = true;
			dataIntervalTrk.keyOverride(sender, e);
		}

		private void slider_ValueChangeEvent(object sender, EventArgs e) {
			if (((PhidgetsControlLibrary.CustomScroll)sender).isFixedRange) //datainterval control
			{
				if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.INTERVAL) {
					try {
						double temp = dataIntervalTrk.Value;
						encoder.DataInterval = (int)temp;
						if (encoder.DataInterval < (int)temp) {
							while (encoder.DataInterval != (int)temp) {
								temp++;
								encoder.DataInterval = (int)temp;
							}
						} else {
							while (encoder.DataInterval != (int)temp) {
								temp--;
								encoder.DataInterval = (int)temp;
							}
						}

						dataIntervalTrk.TrueValue = encoder.DataInterval;
						dataIntervalTrk.Value = encoder.DataInterval;



						updateGraphDataInterval();
					} catch (Exception ex) { errorBox.addMessage("Error setting data interval: " + ex.Message); }
				} else if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.RATE) {
					try {
						encoder.DataRate = dataIntervalTrk.Value;
						updateGraphDataInterval();
					} catch (Exception ex) { errorBox.addMessage("Error setting data rate: " + ex.Message); }
				}
			} else //changetrigger control
			  {
				try {
					encoder.PositionChangeTrigger = (int)changeTriggerTrk.Value;
					updateGraphDataInterval();
				} catch (Exception ex) { errorBox.addMessage("Error setting change trigger: " + ex.Message); }
			}
		}
		private void dataIntervalTrk_ModeSwitch(object sender, EventArgs e) {
			dataIntervalTrk.SwitchModes();
			if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.RATE) {
				try {
					dataIntervalTrk.isIntValue = false;
					dataIntervalTrk.Unit = "Hz";
					dataIntervalTrk.trkText = encoder.DataRate.ToString() + "Hz";
					dataIntervalTrk.Value = encoder.DataRate;
				} catch (Exception ex) { errorBox.addMessage("Error setting data rate: " + ex.Message); }

			} else if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.INTERVAL) {
				try {
					dataIntervalTrk.isIntValue = true;
					if (encoder.DeviceID != DeviceID.PN_1046)
						dataIntervalTrk.Unit = "ms";
					dataIntervalTrk.trkText = encoder.DataInterval.ToString() + "ms";
					dataIntervalTrk.Value = encoder.DataInterval;
				} catch (Exception ex) { errorBox.addMessage("Error setting data interval: " + ex.Message); }
			}
			dataIntervalTrk.TrueValue = dataIntervalTrk.Value;
		}

		private void rotarySelect_CheckedChanged(object sender, EventArgs e) {
			if (calcPositionGraphBox.Visible) {
				calcPositionGraphBox.Close();
			}
			if (rotarySelect.Checked) {
				if (linearSelect.Checked)
					linearSelect.Checked = false;
			}

			cprLbl.Text = "Resolution (CPR):";
			cprTxt.Text = "";
			CPR = -1;
			rpmTxt.Text = "N/A";

			calcPositionGraphBox = null;
			calcPositionGraphBox = new DataGraphBox(this, "Angle");
			calcPositionGraphBox.SetTitle(DataGraphBox.GenerateTitle("Angle", (Phidget)encoder));
			calcPositionLbl.Text = "Angle:";
			calcPositionTxt.Text = "N/A";
			calcPositionLbl.Location = new Point(56, 120);
		}

		private void linearSelect_CheckedChanged(object sender, EventArgs e) {
			if (calcPositionGraphBox.Visible) {
				calcPositionGraphBox.Close();
			}

			if (linearSelect.Checked) {
				if (rotarySelect.Checked)
					rotarySelect.Checked = false;
			}

			cprLbl.Text = "Resolution (µm/pulse):";
			cprTxt.Text = "";
			CPR = -1;
			_rpmTxt = "N/A";

			calcPositionGraphBox = null;
			calcPositionGraphBox = new DataGraphBox(this, "Millimeters");
			calcPositionGraphBox.SetTitle(DataGraphBox.GenerateTitle("Millimeters", (Phidget)encoder));
			calcPositionLbl.Text = "Length:";
			calcPositionTxt.Text = "N/A";
			calcPositionLbl.Location = new Point(50, 120);
		}

		private void velocityGraphButton_Click(object sender, EventArgs e) {
			if (!velocityGraphBox.Visible)
				velocityGraphBox.Show();
		}

		private void calcPositionGraphBtn_Click(object sender, EventArgs e) {
			if (!calcPositionGraphBox.Visible)
				calcPositionGraphBox.Show();
		}
	}
}
